use crate::ip::IpData;
use crate::utils::*;

use chrono::prelude::*;
use chrono::Duration;
use clap::{Arg, ArgMatches, Command};
use nix::sys::inotify::{AddWatchFlags, InitFlags, Inotify, WatchDescriptor};
use regex::Regex;
use reqwest::{Client, Error as ReqError, Response};
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::hash::{Hash, Hasher};
use std::path::Path;

const MASTERSERVER: &str = "ipbl.paulbsd.com";
const ZMQSUBSCRIPTION: &str = "ipbl";

#[derive(Debug, Clone)]
pub struct Context {
    pub blocklist: HashMap<String, IpData>,
    pub cfg: Config,
    pub client: Client,
    pub discovery: Discovery,
    pub flags: Flags,
    pub hostname: String,
    pub instance: Box<Inotify>,
    pub sas: HashMap<String, SetMap>,
    pub hashwd: HashMap<String, WatchDescriptor>,
}

#[derive(Debug, Clone)]
pub struct SetMap {
    pub filename: String,
    pub fullpath: String,
    pub regex: Regex,
    pub set: Set,
    pub watchedfiles: HashMap<String, u64>,
    pub wd: WatchDescriptor,
}

#[derive(Debug, Clone)]
pub struct Flags {
    pub debug: bool,
    pub server: String,
}

impl Context {
    pub async fn new() -> Self {
        // Get flags
        let debug = Context::argparse().is_present("debug");
        let server = Context::argparse()
            .value_of("server")
            .unwrap_or(format!("https://{}", MASTERSERVER).as_str())
            .to_string();

        // Build context
        let mut ctx = Context {
            cfg: Config::new(),
            flags: Flags {
                debug: debug,
                server: server,
            },
            hostname: gethostname(true),
            discovery: Discovery {
                version: "1.0".to_string(),
                urls: HashMap::new(),
            },
            client: Client::builder()
                .user_agent(format!(
                    "{}/{}@{}",
                    env!("CARGO_PKG_NAME"),
                    env!("CARGO_PKG_VERSION"),
                    gethostname(false)
                ))
                .build()
                .unwrap(),
            sas: HashMap::new(),
            instance: Box::new(Inotify::init(InitFlags::empty()).unwrap()),
            blocklist: HashMap::new(),
            hashwd: HashMap::new(),
        };
        ctx.discovery = ctx.discovery().await.unwrap();

        #[cfg(test)]
        return ctx;

        print!("Loading config ... ");
        match ctx.load().await {
            Ok(_) => {}
            Err(err) => {
                println!("error loading config: {err}");
                std::process::exit(1);
            }
        }
        ctx
    }

    pub fn argparse() -> ArgMatches {
        Command::new(env!("CARGO_PKG_NAME"))
            .version(env!("CARGO_PKG_VERSION"))
            .author(env!("CARGO_PKG_AUTHORS"))
            .about(env!("CARGO_PKG_DESCRIPTION"))
            .arg(
                Arg::new("server")
                    .short('s')
                    .long("server")
                    .value_name("server")
                    .default_value(format!("https://{MASTERSERVER}").as_str())
                    .help("Sets a http server")
                    .takes_value(true),
            )
            .arg(
                Arg::new("debug")
                    .short('d')
                    .takes_value(false)
                    .help("Enable debugging"),
            )
            .get_matches()
    }

    pub async fn discovery(&self) -> Result<Discovery, ReqError> {
        let resp: Result<Response, ReqError> = self
            .client
            .get(format!("{server}/discovery", server = self.flags.server))
            .send()
            .await;
        let req = match resp {
            Ok(re) => re,
            Err(err) => return Err(err),
        };
        let data: Discovery = match req.json().await {
            Ok(res) => res,
            Err(err) => return Err(err),
        };
        Ok(data)
    }

    pub async fn load(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        self.cfg.load(self.to_owned()).await?;
        self.create_sas().await?;
        Ok(())
    }

    pub async fn get_blocklist(&self) -> Vec<IpData> {
        let mut res: Vec<IpData> = vec![];
        for (_, v) in self.blocklist.iter() {
            res.push(v.clone());
        }
        res
    }

    pub async fn gc_blocklist(&mut self) -> Vec<IpData> {
        let mut removed: Vec<IpData> = vec![];
        let now: DateTime<Local> = Local::now().trunc_subsecs(0);
        // nightly, future use
        //let drained: HashMap<String,IpData> = ctx.blocklist.drain_filter(|k,v| v.parse_date() < mindate)
        for (id, blocked) in self.blocklist.clone().iter() {
            for set in self.cfg.sets.clone() {
                if blocked.src == set.src {
                    let mindate = now - Duration::minutes(set.blocktime);

                    if blocked.parse_date() < mindate {
                        self.blocklist.remove(&id.to_string()).unwrap();
                        removed.push(blocked.clone());
                    }
                    break;
                }
            }
        }
        removed
    }

    pub async fn update_blocklist(&mut self, ip: &IpData) {
        if !self.blocklist.contains_key(&ip.ip) {
            self.blocklist.insert(ip.ip.clone(), ip.clone());
        }
    }

    pub async fn create_sas(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        for set in &self.cfg.sets {
            let p = Path::new(set.path.as_str());
            if p.is_dir() {
                let res = match self.hashwd.get(&set.path.to_string()) {
                    Some(wd) => *wd,
                    None => {
                        let res = self
                            .instance
                            .add_watch(set.path.as_str(), AddWatchFlags::IN_MODIFY)
                            .unwrap();
                        self.hashwd.insert(set.path.to_string(), res);
                        res
                    }
                };
                let fullpath: String = match set.filename.as_str() {
                    "" => set.path.clone(),
                    _ => {
                        format!(
                            "{path}/{filename}",
                            path = set.path,
                            filename = set.filename.clone()
                        )
                    }
                };
                self.sas.insert(
                    set.src.clone(),
                    SetMap {
                        filename: set.filename.clone(),
                        fullpath: fullpath,
                        set: set.clone(),
                        regex: Regex::new(set.regex.as_str()).unwrap(),
                        wd: res,
                        watchedfiles: HashMap::new(),
                    },
                );
            }
        }
        Ok(())
    }
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct Config {
    pub sets: Vec<Set>,
    #[serde(skip_serializing)]
    pub trustnets: Vec<String>,
    pub zmq: HashMap<String, ZMQ>,
}

impl Config {
    pub fn new() -> Self {
        Self {
            sets: vec![
                Set {
                    src: "smtp".to_string(),
                    filename: "mail.log".to_string(),
                    regex: "(SASL LOGIN authentication failed)".to_string(),
                    path: "/var/log".to_string(),
                    blocktime: 60,
                },
                Set {
                    src: "ssh".to_string(),
                    filename: "auth.log".to_string(),
                    regex: "(Invalid user|BREAK|not allowed because|no matching key exchange method found)".to_string(),
                    path: "/var/log".to_string(),
                    blocktime: 60,
                },
                Set {
                    src: "http".to_string(),
                    filename: "".to_string(),
                    regex: "(anonymousfox.co)".to_string(),
                    path: "/var/log/nginx".to_string(),
                    blocktime: 60,
                }
                ,Set {
                    src: "openvpn".to_string(),
                    filename: "status".to_string(),
                    regex: "(UNDEF)".to_string(),
                    path: "/var/run/openvpn".to_string(),
                    blocktime: 60,
                },
            ],
            trustnets: vec![
                "127.0.0.0/8".to_string(),
                "10.0.0.0/8".to_string(),
                "172.16.0.0/12".to_string(),
                "192.168.0.0/16".to_string(),
            ],
            zmq: HashMap::from([("pubsub".to_string(),ZMQ{
                t: "pubsub".to_string(),
                hostname: MASTERSERVER.to_string(),
                port: 9999,
                subscription: ZMQSUBSCRIPTION.to_string(),
            }),("reqrep".to_string(),ZMQ {
                t: "reqrep".to_string(),
                hostname: MASTERSERVER.to_string(),
                port: 9998,
                subscription: String::new(),
            })])
        }
    }

    pub async fn load(&mut self, ctx: Context) -> Result<(), ReqError> {
        self.get_sets(&ctx).await?;
        self.get_trustnets(&ctx).await?;
        self.get_zmq_config(&ctx).await?;
        Ok(())
    }

    async fn get_trustnets(&mut self, ctx: &Context) -> Result<(), ReqError> {
        let resp: Result<Response, ReqError> = ctx
            .client
            .get(format!(
                "{server}/config/trustlist",
                server = ctx.flags.server
            ))
            .send()
            .await;
        let req = match resp {
            Ok(re) => re,
            Err(err) => return Err(err),
        };
        let data: Vec<String> = match req.json::<Vec<String>>().await {
            Ok(res) => res,
            Err(err) => return Err(err),
        };
        self.trustnets = data;
        Ok(())
    }

    async fn get_sets(&mut self, ctx: &Context) -> Result<(), ReqError> {
        let resp: Result<Response, ReqError> = ctx
            .client
            .get(format!("{server}/config/sets", server = ctx.flags.server))
            .send()
            .await;
        let req = match resp {
            Ok(re) => re,
            Err(err) => return Err(err),
        };
        let data: Vec<Set> = match req.json::<Vec<Set>>().await {
            Ok(res) => res,
            Err(err) => return Err(err),
        };
        self.sets = data;
        Ok(())
    }

    async fn get_zmq_config(&mut self, ctx: &Context) -> Result<(), ReqError> {
        let resp: Result<Response, ReqError> = ctx
            .client
            .get(format!("{server}/config/zmq", server = ctx.flags.server))
            .send()
            .await;
        let req = match resp {
            Ok(re) => re,
            Err(err) => return Err(err),
        };
        let data: HashMap<String, ZMQ> = match req.json::<Vec<ZMQ>>().await {
            Ok(res) => {
                let mut out: HashMap<String, ZMQ> = HashMap::new();
                res.into_iter().map(|x| x).for_each(|x| {
                    out.insert(x.t.to_string(), x);
                });
                out
            }
            Err(err) => return Err(err),
        };
        self.zmq = data;
        Ok(())
    }
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct Set {
    pub src: String,
    pub filename: String,
    pub regex: String,
    pub path: String,
    pub blocktime: i64,
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct ZMQ {
    #[serde(rename = "type")]
    pub t: String,
    pub hostname: String,
    pub port: i64,
    pub subscription: String,
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct Discovery {
    pub version: String,
    pub urls: HashMap<String, URL>,
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct URL {
    pub key: String,
    pub path: String,
}

impl PartialEq for Set {
    fn eq(&self, other: &Self) -> bool {
        self.src == other.src
    }
}

impl Hash for Set {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.src.hash(state);
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::ip::*;
    use Context;

    #[tokio::test]
    pub async fn test_gc_blocklist() {
        let mut ctx = Context::new().await;
        let now: DateTime<Local> = Local::now().trunc_subsecs(0);
        ctx.blocklist = HashMap::new();
        ctx.blocklist.insert(
            "1.1.1.1".to_string(),
            IpData {
                ip: "1.1.1.1".to_string(),
                hostname: "test1".to_string(),
                date: (now - Duration::minutes(61)).to_rfc3339().to_string(),
                src: "ssh".to_string(),
            },
        );
        ctx.blocklist.insert(
            "1.1.1.2".to_string(),
            IpData {
                ip: "1.1.1.2".to_string(),
                hostname: "test2".to_string(),
                date: (now - Duration::minutes(61)).to_rfc3339().to_string(),
                src: "http".to_string(),
            },
        );
        ctx.blocklist.insert(
            "1.1.1.3".to_string(),
            IpData {
                ip: "1.1.1.3".to_string(),
                hostname: "testgood".to_string(),
                date: (now - Duration::minutes(59)).to_rfc3339().to_string(),
                src: "http".to_string(),
            },
        );
        let result = ctx.gc_blocklist().await;
        assert_eq!(result.len(), 2);
    }
}
