use super::*;

use chrono::prelude::*;
use chrono::Duration;
use std::collections::HashMap;
use std::sync::Arc;
use tokio::sync::mpsc::{channel, Receiver, Sender};
use tokio::sync::Mutex;

const BL_CHAN_SIZE: usize = 32;
const ZMQ_CHAN_SIZE: usize = 64;

pub async fn process(ctx: &Arc<Mutex<Context>>) {
    println!(
        "Launching {} version {}",
        env!("CARGO_PKG_NAME"),
        env!("CARGO_PKG_VERSION")
    );

    let (ipdatatx, mut ipdatarx): (Sender<IpData>, Receiver<IpData>) = channel(ZMQ_CHAN_SIZE);

    // initialize the firewall table
    //firewall::init(&env!("CARGO_PKG_NAME").to_string());

    // initialize zeromq sockets
    let reqsocket;
    let subsocket;
    {
        let ctxarc = Arc::clone(&ctx);
        let zmqctx = ctxarc.lock().await;
        reqsocket = zconnect(&zmqctx.cfg.zmq.get("reqrep").unwrap(), zmq::REQ)
            .await
            .unwrap();
        subsocket = zconnect(&zmqctx.cfg.zmq.get("pubsub").unwrap(), zmq::SUB)
            .await
            .unwrap();
    }

    listenpubsub(&ctx, ipdatatx.clone(), subsocket).await;

    let mut blrx = watchfiles(&ctx).await;

    let ctxarc = Arc::clone(&ctx);
    tokio::spawn(async move {
        compare_files_changes(&ctxarc, &mut blrx, &ipdatatx).await;
    });

    loop {
        let mut ret: Vec<String> = Vec::new();
        let begin: DateTime<Local> = Local::now().trunc_subsecs(0);

        // wait for logs parse and zmq channel receive
        let ip = ipdatarx.recv().await.unwrap();

        // lock the context mutex
        let ctxarc = Arc::clone(&ctx);
        let mut ctx = ctxarc.lock().await;

        // refresh context blocklist
        ctx.update_blocklist(&ip).await;
        ctx.gc_blocklist().await;

        // send ip list to ws and zmq sockets
        if ip.hostname == ctx.hostname {
            send_to_ipbl_ws(&ctx, &ip, &mut ret).await;
            send_to_ipbl_zmq(&reqsocket, &ip).await;
        }

        // apply firewall blocking
        firewall::block(
            &env!("CARGO_PKG_NAME").to_string(),
            &ctx.get_blocklist().await,
            &mut ret,
        )
        .unwrap();

        // log lines
        println!("{ret}", ret = ret.join(", "));

        let end: DateTime<Local> = Local::now().trunc_subsecs(0);
        if (end - begin) > Duration::seconds(5) {
            // reload configuration from the server
            match ctx.load().await {
                Ok(_) => {}
                Err(err) => {
                    println!("error loading config: {err}");
                }
            }
        }
    }
}

async fn watchfiles(ctx: &Arc<Mutex<Context>>) -> Receiver<FileEvent> {
    let (bltx, blrx): (Sender<FileEvent>, Receiver<FileEvent>) = channel(BL_CHAN_SIZE);
    let ctx = Arc::clone(ctx);
    tokio::spawn(async move {
        loop {
            let events: Vec<InotifyEvent>;
            {
                let ctx = ctx.lock().await;
                events = ctx.instance.read_events().unwrap();
            }

            for event in events {
                let date: DateTime<Local> = Local::now().trunc_subsecs(0);
                bltx.send(FileEvent {
                    ie: event,
                    date: date,
                })
                .await
                .unwrap();
            }
        }
    });
    blrx
}

async fn get_last_file_size(watchedfiles: &mut HashMap<String, u64>, path: &str) -> u64 {
    let currentlen = match std::fs::metadata(&path.to_string()) {
        Ok(u) => u.len().clone(),
        Err(_) => 0u64,
    };
    let lastlen = match watchedfiles.insert(path.to_string(), currentlen) {
        Some(u) => u,
        None => 0,
    };
    lastlen
}

async fn compare_files_changes(
    ctx: &Arc<Mutex<Context>>,
    inotifyrx: &mut Receiver<FileEvent>,
    ipdatatx: &Sender<IpData>,
) {
    let mut trustnets;
    loop {
        let modifiedfiles = inotifyrx.recv().await.unwrap();
        let mut list: Vec<IpData> = vec![];

        let mut ctx = ctx.lock().await;
        trustnets = build_trustnets(&ctx.cfg.trustnets);

        match modifiedfiles.ie.name {
            Some(name) => {
                let inotify_filename = name.to_str().unwrap();
                for sak in &mut ctx.clone().sas.keys() {
                    let sa = &mut ctx.sas.get_mut(sak).unwrap();
                    if modifiedfiles.ie.wd == sa.wd {
                        let handle_filename: String;
                        if sa.filename.as_str() == "" {
                            handle_filename = format!("{}/{}", &sa.fullpath, inotify_filename);
                        } else if inotify_filename.starts_with(sa.filename.as_str()) {
                            handle_filename = sa.fullpath.to_owned();
                        } else {
                            continue;
                        }

                        let filesize =
                            get_last_file_size(&mut sa.watchedfiles, &handle_filename).await;

                        match read_lines(&handle_filename, filesize) {
                            Some(lines) => {
                                filter(
                                    lines,
                                    &mut list,
                                    &trustnets,
                                    &sa.regex,
                                    &sa.set.src,
                                    &modifiedfiles.date,
                                );
                            }
                            None => {}
                        };
                        break;
                    }
                }
                drop(ctx);
                for ip in list {
                    ipdatatx.send(ip).await.unwrap();
                }
            }
            None => {}
        }
    }
}
